// params.cc

#include "params.h"
#include <iostream>
#include <ctime>

namespace CommonTilton
{

 // Constructor
  Params::Params(const string& value)
  {
    version = value;

   // Flags for parameters under user control.
    ncols_flag = false;
    nrows_flag = false;
    nbands_flag = false;
    object_labels_map_flag = false;
    region_objects_flag = false;
    region_sum_flag = false;
    region_std_dev_flag = false;
    region_boundary_npix_flag = false;
    region_threshold_flag = false;
//    region_nghbrs_list_flag = false; // Always assumed true!
    region_nghbrs_list_flag = true;
    region_nb_objects_flag = false;
    region_objects_list_flag = false;

   // Flags for internal program parameters (NOT under user control).
    region_sumsq_flag = false;
    region_sumxlogx_flag = false;

    return;
  }

 // Destructor...
  Params::~Params() 
  {
    return;
  }

 // Print version
  void Params::print_version( )
  {
    cout << endl << version << endl << endl;

#if (defined(WINDOWS) || defined(CYGWIN))
    cout << endl << "Copyright (C) 2006 United States Government as represented by the" << endl;
#else
    cout << endl << "Copyright \u00a9 2006 United States Government as represented by the" << endl;
#endif
    cout << "Administrator of the National Aeronautics and Space Administration." << endl;
    cout << "No copyright is claimed in the United States under Title 17, U.S. Code." << endl;
    cout << "All Other Rights Reserved." << endl << endl;
  }

 // Read initial parameters
  bool Params::read_init(const char *param_file)
  {
//    cout << "Reading initial parameters from " << param_file << endl;
    ifstream param_fs(param_file);
    if (!param_fs.is_open())
    {
      cout << "Cannot open input file " << param_file << endl;
      return false;
    }

  // Read initial parameters from parameter file until End-of-File reached
    int sub_pos;
    string line, sub_string;

    while (!param_fs.eof())
    {
      getline(param_fs,line);
      sub_pos = line.find("-");
      while ((!param_fs.eof()) && (sub_pos != 0))
      {
        getline(param_fs,line);
        sub_pos = line.find("-");
      }
      if (line.find("-ncols") != string::npos)
      {
        sub_string = process_line(line,false);
        ncols = atoi(sub_string.c_str());
        ncols_flag = true;
      }
      if (line.find("-nrows") != string::npos)
      {
        sub_string = process_line(line,false);
        nrows = atoi(sub_string.c_str());
        nrows_flag = true;
      }
      if (line.find("-nbands") != string::npos)
      {
        sub_string = process_line(line,false);
        nbands = atoi(sub_string.c_str());
        nbands_flag = true;
      }
    }

    param_fs.close();

// Exit with false status if a required parameter is not set, or an improper parameter
// setting is detected.
    if (!ncols_flag)
    {
      cout << "ERROR: -ncols (Number of columns in input image data) is required" << endl;
      return false;
    }
    if (!nrows_flag)
    {
      cout << "ERROR: -nrows (Number of rows in input image data) is required" << endl;
      return false;
    }
    if (!nbands_flag)
    {
      cout << "ERROR: -nbands (Number of spectral bands in input image data) is required" << endl;
      return false;
    }
    
    return true;
  } // bool Params::read_init(const char *param_file)

 // Read remaining parameters
  bool Params::read(const char *param_file)
  {
//    cout << "Reading remaining parameters from " << param_file << endl;
    ifstream param_fs(param_file);
    if (!param_fs)
    {
      cout << "Cannot open input file " << param_file << endl;
      return false;
    }

  // Read the rest of the parameters from parameter file until End-of-File reached
    int band;
    int sub_pos;
    string line, sub_string;
    while (!param_fs.eof())
    {
      getline(param_fs,line);
      sub_pos = line.find("-");
      while ((!param_fs.eof()) && (sub_pos != 0))
      {
        getline(param_fs,line);
        sub_pos = line.find("-");
      }
      if (line.find("-scale") != string::npos)
      {
        line = process_line(line,true);
        band = 0;
        while (line.size() > 0)
        {
          if (band > nbands)
            break;
          sub_pos = line.find_first_of(",");
          if (sub_pos != (int) string::npos)
          {
            sub_string = line.substr(0,sub_pos);
            line = line.substr(sub_pos+1);
          }
          else
          {
            sub_string = line;
            line = "";
          }
          scale[band] = atof(sub_string.c_str());
          band++;
        }
      }
      if (line.find("-offset") != string::npos)
      {
        line = process_line(line,true);
        band = 0;
        while (line.size() > 0)
        {
          if (band > nbands)
            break;
          sub_pos = line.find_first_of(",");
          if (sub_pos != (int) string::npos)
          {
            sub_string = line.substr(0,sub_pos);
            line = line.substr(sub_pos+1);
          }
          else
          {
            sub_string = line;
            line = "";
          }
          offset[band] = atof(sub_string.c_str());
          band++;
        }
      }
      if (line.find("-class_labels_map") != string::npos)
      {
        sub_string = process_line(line,false);
        class_labels_map_file = sub_string;
      }
      if (line.find("-region_classes") != string::npos)
      {
        region_classes_file = process_line(line,false);
      }
      if (line.find("-object_labels_map") != string::npos)
      {
        sub_string = process_line(line,false);
        object_labels_map_file = sub_string;
        object_labels_map_flag = true;
      }
      if (line.find("-region_objects") != string::npos)
      {
        if (line.find("-region_objects_list") != string::npos)
        {
          sub_string = process_line(line,false);
          sub_pos = atoi(sub_string.c_str());
          if (sub_pos == 1)
            region_objects_list_flag = true;
          else
            region_objects_list_flag = false;
        }
        else
        {
          region_objects_file = process_line(line,false);
          region_objects_flag = true;
        }
      }
      if (line.find("-region_sum") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_sum_flag = true;
        else
          region_sum_flag = false;
      }
      if (line.find("-region_std_dev") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_std_dev_flag = true;
        else
          region_std_dev_flag = false;
      }
      if (line.find("-region_boundary_npix") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_boundary_npix_flag = true;
        else
          region_boundary_npix_flag = false;
      }
      if (line.find("-region_threshold") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_threshold_flag = true;
        else
          region_threshold_flag = false;
      }
/*
      if (line.find("-region_nghbrs_list") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_nghbrs_list_flag = true;
        else
          region_nghbrs_list_flag = false;
      } 
*/
      if (line.find("-region_nb_objects") != string::npos)
      {
        sub_string = process_line(line,false);
        sub_pos = atoi(sub_string.c_str());
        if (sub_pos == 1)
          region_nb_objects_flag = true;
        else
          region_nb_objects_flag = false;
      }
    }

    param_fs.close();

    bool status = true;

    return status;
  }  // bool Params::read(const char *param_file)

 // oParams Constructor
  oParams::oParams()
  {
    return;
  }

 // oParams Destructor...
  oParams::~oParams() 
  {
    return;
  }

 // Print parameters
  void Params::print()
  {
      int band;

  // Print version
      cout << "This is " << version << endl << endl;

  // Print input parameters
      cout << endl << "Parameters in Params class:" << endl;
      cout << "Input image number of columns = " <<  ncols << endl;
      cout << "Input image number of rows = " <<  nrows << endl;
      cout << "Input image number of bands = " <<  nbands << endl;
      if (nbands < 10)
      {
        cout << "Input data scale factors: ";
        for (band = 0; band < nbands; band++)
          cout << scale[band] << " ";
        cout << endl;
        cout << "Input data offset factors: ";
        for (band = 0; band < nbands; band++)
          cout << offset[band] << " ";
        cout << endl;
      }
      cout << "Output class labels map file name: " << class_labels_map_file << endl;
      cout << "Output region classes file name: " << region_classes_file << endl;
      if (object_labels_map_flag)
        cout << "Output object labels map file name: " << object_labels_map_file << endl;
      if (region_objects_flag)
        cout << "Output region objects file name: " << region_objects_file << endl;
      if (region_sum_flag)
        cout << "Region sum (and, if available, sumsq and sumxlogx) information included." << endl;
      if (region_std_dev_flag)
        cout << "Region standard deviation information included" << endl;
      if (region_sumsq_flag)
        cout << "Region sumsq information included." << endl;
      if (region_sumxlogx_flag)
        cout << "Region sumxlogx information included." << endl;
      if (region_boundary_npix_flag)
        cout << "Region region boundary number of pixels information included" << endl;
      if (region_threshold_flag)
        cout << "Region most recent merge threshold information included" << endl;
      if (region_nghbrs_list_flag)
        cout << "List of region clases neighboring each region class included" << endl;
      if (region_nb_objects_flag)
        cout << "Number of region objects contained in each region class included" << endl;
      if (region_objects_list_flag)
        cout << "List of region objects contained in each region class included" << endl;

  }

  string process_line(const string& line, const bool& list_flag)
  {
    int sub_pos, sub_pos_space, sub_pos_tab;
    string sub_string;

    sub_pos_space = line.find_first_of(" ");
    sub_pos_tab = line.find_first_of("\t");
    sub_pos = sub_pos_space;
    if (sub_pos ==  (int) string::npos)
      sub_pos = sub_pos_tab;
    if (sub_pos ==  (int) string::npos)
      return " ";

    sub_string = line.substr(sub_pos);
    while ((sub_string.substr(0,1) == "\t") || (sub_string.substr(0,1) == " "))
      sub_string = line.substr(++sub_pos);
#ifndef WINDOWS
    if (list_flag)
      return sub_string;

    sub_pos = sub_string.find_first_of(" ");
    if (sub_pos !=  (int) string::npos)
      sub_string = sub_string.substr(0,sub_pos);
    sub_pos = sub_string.find_first_of("\t");
    if (sub_pos !=  (int) string::npos)
      sub_string = sub_string.substr(0,sub_pos);
    sub_pos = sub_string.find_first_of("\r");
    if (sub_pos !=  (int) string::npos)
      sub_string = sub_string.substr(0,sub_pos);
#endif
    return sub_string;
  }

} // namespace CommonTilton
