/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<<
   >>>>
   >>>>       Purpose:  Provides the Params class, which is a class
   >>>>                 holding program parameters.
   >>>>
   >>>>    Written By:  James C. Tilton, MC 606.3, NASA's GSFC, Greenbelt, MD 20771
   >>>>                 e-mail:  James.C.Tilton@nasa.gov
   >>>>          Date:  June 21, 2016 - Based on Params class from RHSEG/rhsegV1.64
   >>>> Modifications:  
   >>>>
   >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> <<<<<<<<<<<<<<<<<<<<<<<<<< */
#ifndef PARAMS_H
#define PARAMS_H

#include <string>
#include <fstream>
#include <sstream>
#include <cstdlib>
#include <climits>
#include <stdexcept>
#include <vector>
#include <gdal_priv.h>

using namespace std;

namespace CommonTilton
{
  class oParams;

// Params class
  class Params
  {
    public:
    // Constructor and Destructor
      Params(const string& value);
      virtual ~Params();

    // Member functions
      void print_version();
      bool read_init(const char *param_file);
      bool read(const char *param_file);
      void print();

    // Member variables (public)
    /*-- Number of columns in input image data file --*/
      int    ncols;                /*-- PROGRAM PARAMETER --*/
      bool   ncols_flag;           /*-- EXISTENCE FLAG --*/

    /*-- Number of rows in input image data file --*/
      int    nrows;                /*-- PROGRAM PARAMETER --*/
      bool   nrows_flag;           /*-- EXISTENCE FLAG --*/

    /*-- Number of spectral bands in input image data file --*/
      int    nbands;               /*-- PROGRAM PARAMETER --*/
      bool   nbands_flag;          /*-- EXISTENCE FLAG --*/

    /*-- Input image data scale and offset values (optional) --*/
      vector<double> scale, offset;     /*-- USER INPUT PARAMETERS --*/

    /*-- Output region classes map file (required) --*/
      string  class_labels_map_file;    /*-- USER INPUT FILENAME --*/

    /*-- Output region object list file (required) --*/
      string region_classes_file;       /*-- USER INPUT FILENAME --*/

    /*-- Output region objects map file (optional) --*/
      string object_labels_map_file;    /*-- USER INPUT FILENAME --*/
      bool   object_labels_map_flag;    /*-- EXISTENCE FLAG --*/

    /*-- Output region objects file (optional) --*/
      string region_objects_file;       /*-- USER INPUT FILENAME --*/
      bool   region_objects_flag;       /*-- EXISTENCE FLAG --*/

    /*-- Region sum output flag --*/
      bool   region_sum_flag;           /*-- USER INPUT PARAMETER --*/

    /*-- Region standard deviation output flag --*/
      bool   region_std_dev_flag;       /*-- USER INPUT PARAMETER --*/

    /*-- Flag that signifies whether or not region sumsq information is required --*/
      bool   region_sumsq_flag;    /*-- PROGRAM PARAMETER --*/

    /*-- Flag that signifies whether or not region sumxlogx information is required --*/
      bool   region_sumxlogx_flag; /*-- PROGRAM PARAMETER --*/

    /*-- Region boundary number of pixels output flag --*/
      bool   region_boundary_npix_flag; /*-- USER INPUT PARAMETER --*/

    /*-- Region merge threshold output flag --*/
      bool   region_threshold_flag;     /*-- USER INPUT PARAMETER --*/

    /*-- Output flag of the list of region classes neighboring each region class --*/
      bool   region_nghbrs_list_flag;   /*-- USER INPUT PARAMETER --*/

    /*-- Region number of regions objects flag --*/
      bool   region_nb_objects_flag;    /*-- USER INPUT PARAMETER --*/

    /*-- List of region objects output flag --*/
      bool   region_objects_list_flag;  /*-- USER INPUT PARAMETER --*/

    /* Program version */
      string version;              /* -- PROGRAM PARAMETER --*/

    /*-- Driver for output image files --*/
      GDALDriver *driver;          /*-- PROGRAM PARAMETER --*/

    // FRIEND FUNCTIONS and CLASSES
    protected:

    private:
  };

  // For "stringify'ing integer and float values
  class BadConversion : public std::runtime_error
  {
    public:
      BadConversion(const std::string& s)
        : std::runtime_error(s)
        { }
  };

  inline std::string stringify_int(int x)
  {
    std::ostringstream o;
    if (!(o << x))
      throw BadConversion("stringify_int(int)");
    return o.str();
  }

  inline std::string stringify_float(float x)
  {
    std::ostringstream o;
    if (!(o << x))
      throw BadConversion("stringify_float(float)");
    return o.str();
  }

// Other associated classes and functions
  class oParams
  {
    public:
    // Constructor and Destructor
      oParams();
      virtual ~oParams();

    // Member variables (public)
      vector<double> scale, offset;
      int nb_levels, level0_nb_classes, level0_nb_objects;
      vector<unsigned int> int_buffer_size;
  };

  string process_line(const string& line, const bool& list_flag);

} // CommonTilton

#endif /* PARAMS_H */
